/////////////////////////////////////////////////////////////////////
//                       FASTARI.CPP v.1.0
//
//  Nania Francesco Antonio  (Italy Mer (ME))
// (C) 2006, under GPL, http://www.gnu.org/licenses/gpl.txt
//
// DESCRIPTION
// >>>>>>>>>>>
//
//  This program is a order-2 adaptive arithmetic packer .
//
// Compilation instructions:
// >>>>>>>>>>>>>>>>>>>>>>>>>
//
//  Borland C++     : bcc32 -w fastari.cpp
//  g++             : g++ -o fastari ari.cpp
//
//  To Use
//  >>>>>>
//
//   Fastari c  inputfile outputfile --> compressed-file            //
//                                                                  //
//////////////////////////////////////////////////////////////////////
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <cassert>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctime>
#if !defined( unix )
#include <io.h>
#endif
// Variables Declarations
//////////////////////////////////////////////////////////////
FILE *ein;
FILE *eout;
unsigned long rc,r1,r2;
unsigned short EOFS=257;
unsigned short byteindex[65536][256];
unsigned short indexbyte[65536][258];
unsigned short freq[65536][258];      // Symbol frequencies
unsigned short sfreq[65536][258];  // Cumulative symbol frequencies
unsigned char Mbits=16 ;             // Number of bits in a code value
long Maxvalue=65535; // range of sum of frequencies
unsigned short qtr=((Maxvalue>>2)+1);       // 1 quarter
unsigned short qtr2=(2*qtr);              // 2 quarter
unsigned short qtr3=(3*qtr);         //3 quarter
unsigned short Maxf= 16383;            // Maximum freq. count = ((1<<16)/4)-1
int writer=0;
int bitsenc=8;
long low, high,folder;
long bitsf;
int extbits;
//////////////////////////////////////////////////////////////////////////////////777

void startenc( void )
{
    for (int old = 0; old<=65535; old++)
    {
    for (int i = 0; i<256; i++)
    {
    indexbyte[old][i+1] = i;
    byteindex[old][i] = i+1;
    }
    for (int i = 0; i<=257; i++)
    {
    sfreq[old][i] = 257-i;
    freq[old][i] = 1;
    }
    freq[old][0] = 0;
    }
}

inline void outbit( int bit )
{   writer >>= 1;
    if (bit==1) writer |= 128;
    bitsenc --;
    if (bitsenc==0) {bitsenc = 8; putc( (char) writer, eout );}
}
inline int inbit( void )
{   int bit;
    if (bitsenc==0) {
    writer = getc(ein);      // read byte
    if (writer==EOF) {
    extbits ++;
    if (extbits>Mbits-2)
    {
    fprintf(stderr,"Error in input file\n");
    exit(-1);
    }
    }
    bitsenc = 8;
    }
    bit = (writer&1);
    writer >>= 1;
    bitsenc --;
    return bit;
}

void bitplus( int bit )
{   outbit(bit);
    while (bitsf>0)
    {outbit(1-bit); bitsf --;}
}


// Encode/decode byte or symbol
//////////////////////////////////////////////////////
int encode(int esymbol )
{   long range;
    int sfre;
    int rsymbol;
    range = (long)(high-low)+1;
    high=low;
    rsymbol=esymbol;
    if (rsymbol<0)
    {
    sfre = (int)((((long)(folder-low)+1)*sfreq[rc][0]-1)/range);
    for (esymbol = 1; sfreq[rc][esymbol]>sfre; esymbol++) ;
    }
    low  +=(range*sfreq[rc][esymbol])/sfreq[rc][0];
    high +=(range*sfreq[rc][esymbol-1])/sfreq[rc][0]-1;
    for (;;)
    {
        if (high<qtr2) {if (rsymbol>0) bitplus(0); } // in encode mode
        else if (low>=qtr2)
        {
        if (rsymbol>0) bitplus(1);     // in encode mode
        if (rsymbol<0) folder -= qtr2; // in decode mode
        low  -= qtr2; high -= qtr2;
        }
        else if (low>=qtr && high<qtr3)
        {
        if (rsymbol>0) bitsf ++;        // in encode mode
        if (rsymbol<0) folder -= qtr;   // in decode mode
        low -= qtr;   high -= qtr;
        }
        else break;
        low = (low<<1);
        high = (high<<1)+1;
        if (rsymbol<0) folder = (folder<<1)+inbit(); // in decode mode
    }
    if (rsymbol<0) return esymbol;                   // in decode mode
}


// UPDATE THE MODEL

void update( int ich )
{   int i;
    int sum=0;
    int tch;
    int tchsymbol;
    if (sfreq[rc][0]==Maxf)
    {
        for (i = 257; i>=0; i--)
        {
           ++freq[rc][i]>>=1;
           sfreq[rc][i] = sum;
           sum+= freq[rc][i];
        }
    }
    i=ich;
    while(freq[rc][i]==freq[rc][i-1])
    i--;
    freq[rc][i]++;
    if (i<ich)
    {
    tch = indexbyte[rc][i];
    tchsymbol = indexbyte[rc][ich];
    indexbyte[rc][ich] = (unsigned char) tch;
    indexbyte[rc][i] = (unsigned char) tchsymbol;
    byteindex[rc][tch] = ich;
    byteindex[rc][tchsymbol] = i;
    }
    if (i>0)
    for (int e = i-1; e>=0; e--)
    sfreq[rc][e] ++;

}


//////////////////////////// main ////////////////////////////

int main(int argc, char** argv) {

  // Chech arguments: fastari c/d input output
  if (argc!=4 || (argv[1][0]!='c' && argv[1][0]!='d')) {
    printf("                              FASTARI v1.0                                    \n");
    printf("Fast aritmetic compressor order2 by Nania Francesco Antonio (Italy) 2006      \n");
    printf("\n");
    printf("To pack   file: fastari c input output\n"
           "To unpack file: fastari d input output\n");
    exit(1);
  }

  // Start timer
  clock_t start = clock();

  // Open files
  FILE *in=fopen(argv[2], "rb");
  if (!in) perror(argv[2]), exit(1);
  FILE *out=fopen(argv[3], "wb");
  if (!out) perror(argv[3]), exit(1);
  eout=out;
  ein=in;


  if (argv[1][0]=='c') // Pack
  {
    clock_t start = clock();
    startenc();
    low = 0;
    high = Maxvalue;
    bitsf = 0;
    for (;;) {
        int ch;
        int ich;
        ch = getc(in);
        if (ch==EOF) break;                     // end of loop
        ich = byteindex[rc][ch];
        encode(ich);
        update(ich);
        r2=r1;
        r1=ch;
        rc=r1+(r2<<8);
    }
    encode(EOFS);         // Encode EOF symbol.  //
    bitsf ++;
    if (low<qtr) bitplus(0);
    else bitplus(1);
    putc( (char) ( writer >> bitsenc ),eout);
  }
  else  // Unpack
  {
    startenc();
    bitsenc = 0;
    extbits = 0;
    folder = 0;
    for (int i = 1; i<=Mbits; i++) {
    folder = (folder<<1)+inbit();
    }
    low = 0;
    high = Maxvalue;
    for (;;) {
        int ch;
        int ich;
        ich=encode(-1);
        if (ich==EOFS) break;       // end of loop if  decode EOF symbol
        ch = indexbyte[rc][ich];
        putc((char) ch,out);                 // Write byte.
        update(ich);
        r2=r1;
        r1=ch;
        rc=r1+(r2<<8);
  }

}

  // Print results

  printf("%s (%ld bytes) -> %s (%ld bytes) in %1.2f s.\n",argv[2],
ftell(ein), argv[3], ftell(eout),((double)clock()-start)/CLOCKS_PER_SEC);
  return 0;
}
